////////////////////////////////////////////////////////////////////////////////
// File         : Hnet.cpp
// Subsystem    : Graph
// Component    : H-net
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class Hnet
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 10.09.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////

#include "Hnet.h"
#include "HnetUniverse.h"

static IsomorphyID currentIsomophyID = 0; // delete later

IsomorphyID GetNextIsomophyID( void ) { return currentIsomophyID++; }

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Hnet::Hnet( HnetUniverse *universe) : m_universe( universe ), m_virgin( false ), m_elementary( false )
{
	if ( universe != NULL)
		m_isoType[ Root() ] = universe->Root();
}


Hnet::Hnet( HnetUniverse *universe, IsoTypeNode &isoType )
	:  m_universe( universe ), m_virgin( false ), m_elementary( false )
{
	m_isoType[ Root() ] = isoType;
}


Hnet::Hnet( int numOuterEntries, int numOuterExits, HnetUniverse *universe, IsoTypeNode &isoType, bool elementary )
	:  m_universe( universe ), m_virgin( false ), m_elementary( elementary )
{
	int i;

	for ( i=0; i < numOuterEntries; i++)
		AddOuterEntry();

	for ( i=0; i < numOuterExits; i++)
		AddOuterExit();
	
	m_isoType[ Root() ] = isoType;
}


Hnet::Hnet( const Hnet &hnet )
{
	Copy( hnet );
}

	
Hnet::~Hnet()
{

}


//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

Hnet& Hnet::operator=( const Hnet &hnet)
{
	if (this != &hnet)
		Copy( hnet );

	return *this;
}



Node Hnet::AddSubcentre( const IsoTypeNode &isoType )
{ 
	CheckID(); 

	Hnet &hnet = m_universe->GetHnet( isoType );
	Node s = m_Snet.CopyAsSubcentre( hnet.m_Snet ); 
	m_isoType[ s ] = isoType;
	m_universe->AddReference( GetIsoType(), isoType );
	return s; 
}


Node Hnet::CopySubcentre( const Node &s )
{ 
	CheckID(); 

	Node ss = m_Snet.CopySubcentre( s ); 
	m_isoType[ ss ] = m_isoType[ s ];
	m_universe->AddReference( GetIsoType(), m_isoType[ s ] );
	return ss; 
}


void Hnet::DelSubcentre( Node &s  )
{ 
	IsoTypeNode &isoType = m_isoType[ s ];

	CheckID(); 
	m_Snet.DelSubcentre( s ); 
	m_universe->DelReference( GetIsoType(), isoType );
}


void Hnet::ChangeIsoType( const Node &s, const IsoTypeNode &isoType )
{
	if ( m_isoType[ s ] == isoType )
		return; // nothing to do

	CheckID(); 
	Hnet &hnet =  m_universe->GetHnet( isoType );
	// Attention! outer entries are inner exits, outer exits are inner entries
	m_Snet.ChangeSubcentre( s, hnet.NumberOfOuterExits(), hnet.NumberOfOuterEntries() );

	IsoTypeNode &isoTypeOld = m_isoType[ s ];
	m_isoType[ s ] = isoType;

	m_universe->DelReference( GetIsoType(), isoTypeOld );
	m_universe->AddReference( GetIsoType(), isoType );
}


bool Hnet::Elementary( Node &s ) 
{ 
	return (m_universe->GetHnet( m_isoType[ s ] )).m_elementary; 
}


Node Hnet::Expand( Node &s )
{
	IsoTypeNode isoType = m_isoType[ s ];
	Hnet &hnet =  m_universe->GetHnet( isoType );
	if ( hnet.Elementary() )
		return s;  // is not allowed to expand

	CheckID(); 
	Node sOld = hnet.FirstSubcentre();
	Node sRet, sNew = m_Snet.Replace( s, hnet.m_Snet );
	sRet = sNew;

	// copy isoTypes
	while ( sOld && sNew )
	{
		IsoTypeNode iso = m_isoType[ sNew ] = hnet.m_isoType[ sOld ];
		m_universe->AddReference( GetIsoType(), iso );
		sOld = hnet.NextSubcentre( sOld );
		sNew = NextSubcentre( sNew );
	}

	m_universe->DelReference( GetIsoType(), isoType );
	return sRet;
}



Node Hnet::ExpandAll( void )
{
	Node sn, s = LastSubcentre();

	while ( s )
	{
		sn = PrevSubcentre( s );
		Expand( s );
		s = sn;
	}

	return FirstSubcentre();
}



bool Hnet::Flat( void )
{
	Node s = FirstSubcentre();

	while ( s )
	{
		if ( !Elementary( s ) )
			return false;

		s = NextSubcentre( s );
	}

	return true;
}


int Hnet::NumberOfReferences(  const IsoTypeNode &isoType ) const
{
	int refCount = 0;
	Node s = FirstSubcentre();

	while ( s )
	{
		if ( m_isoType[ s ] == isoType)
			refCount++;

		s = NextSubcentre( s );
	}

	return refCount;
}



//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

// copy without information to H-Net universe; 1:1 copy

void Hnet::Copy( const Hnet &hnet )
{
	Node s, sOrig;
	m_Snet = hnet.m_Snet;
	m_isoID = hnet.m_isoID; 
	m_elementary = hnet.m_elementary;
	m_universe = hnet.m_universe; 
	m_virgin = hnet.m_virgin;

	// copy isomorphy types
	m_isoType[ Root() ] = hnet.m_isoType[ hnet.Root() ];

	s = FirstSubcentre();
	sOrig = hnet.FirstSubcentre();

	while ( s && sOrig )
	{
		m_isoType[ s ] = hnet.m_isoType[ sOrig ];
		s = NextSubcentre( s );
		sOrig = hnet.NextSubcentre( sOrig );
	}
}


//////////////////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////////////////

void Hnet::CheckID( void )
{
	if (m_virgin)
	{
		m_isoID = GetNextIsomophyID();  //first change
		m_virgin = false;
		m_elementary = false;
	}
}
