////////////////////////////////////////////////////////////////////////////////
// File         : graph.h
// Subsystem    : Graph
// Component    : Graph
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class Hnet
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 10.09.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////

#ifndef GRAPH_H
#define GRAPH_H


#ifdef WIN32
#pragma warning( disable : 4786 )  // Disable warning message C4786
#endif

#include <vector>
#include <list>
#include <deque>

#include "indexattr.h"
#include "SUVector.h"

#ifdef WIN32
#define BRUTAL_CAST // must be set for Win32
#else
#define BRUTAL_CAST // can be set for Gnu
#endif


class Node;
class Edge;

// ****************************** Graph ************************************

class Graph
{

public:
	Graph() {};
	Graph(const Graph &g);
	Graph& operator=(const Graph &g);
	int NumberOfNodes(void) const;
	int NumberOfEdges(void) const;
	int NumberOfIngoingEdges( const Node &n) const;
	int NumberOfOutgoingEdges( const Node &n) const;
	Node CreateNode(void);	// append a new node to end of list of nodes
	void Delete(Node &n);	
	Edge CreateEdge(const Node &s, const Node &t);	// append a new edge to end of list of edges; do not check, if edge already exists!!!!!!!!!!!!!!!! 
	Edge CreateEdgeSingle(const Node &s, const Node &t); // like CreateEdge(), but not appendet if edge with this nodes exists already
	void Delete(Edge &e);	
	Node FirstNode(void) const;
	Edge FirstEdge(void) const;
	Node LastNode(void) const;
	Edge LastEdge(void) const;
	Node AddCopyOf(const Graph &g);
	void DeleteAllNodes(void);
	Edge ExistingEdge( const Node &n1, const Node &n2 ) const;
	Edge AddTransitiveHull( const Node &n1, const Node &n2 ); //add all edges from predecestors of n1 to ancestors of n2
	Edge AddTransitiveHull( const Node &n ); // { return AddTransitiveHull( n, n ); } ;
	Edge AddTransitiveHull( const Edge &e );	// { return AddTransitiveHull( e.SourceNode(), e.TargetNode() ); };
	void DeleteTransitiveHull( const Node &n1, const Node &n2 ); //delete all edges from predecestors of n1 to ancestors of n2
	void DeleteTransitiveHull( const Node &n ); // { DeleteTransitiveHull( n, n ); } ;
	void DeleteTransitiveHull( Edge &edge );	// {  DeleteTransitiveHull( e.SourceNode(), e.TargetNode() ); };
	Edge DeleteTransitiveHullReverse( Edge &edge );	// { similar to DeleteTransitiveHull( e.TargetNode(), e.SourceNode() ); };
	Node AddStar( const Node &n1, const Node &n2 ); //add a node and connet to all predecestors of n1 and to all ancestors of n2
	Node AddStar( const Edge &e );	// { return AddStarBetween( e.SourceNode(), e.TargetNode() ); };
	Node AddStarBetween( const Node &n ); //create new predecestors and ancestors an connect to old ones
	void DeleteStarBetween( const Node &n1, const Node &n2 ); //delete all predecestors/ancessors and connect directly to n1/n2 
	void DeleteStarBetween( const Node &n ); //delete all predecestors and ancessors and connect directly to n
	void DeleteStarBetween( const Edge &e );	// { DeleteStarBetween( e.SourceNode(), e.TargetNode() ); };
	Node CreateNodeCopy( const Node &n );	// add Node and copy edges
	void IsolateNode( const Node &n );	// del ingoing and outgoing edges


	// friend classes
	friend class Node;
	friend class Edge;

private:
	// linking informations for nodes and edges 
	struct NodeData;
	struct EdgeData;
	struct NeighbourData;

	
#ifdef BRUTAL_CAST  // neccessary in case of WIN32 and lists

	typedef void **EdgeDataPtr;
	typedef void **NodeDataPtr;

	struct NeighbourData
	{
		NodeDataPtr	_n;
		EdgeDataPtr	_e;
	};

#endif //#ifdef BRUTAL_CAST

	struct NodeData
	{
		std::list<AttributeVector>::iterator attr;
		std::list<std::list<NeighbourData> >::iterator	_inNeighbours, _outNeighbours;
	};

	typedef std::list<NodeData>::iterator NodeDataRef;

	struct EdgeData
	{
		std::list<AttributeVector>::iterator attr;
		std::list<std::list<NeighbourData> >::iterator	_inNeighbours, _outNeighbours;
		std::list<NeighbourData>::iterator		_sourceNeighbour, _targetNeighbour;
	};

	typedef std::list<EdgeData>::iterator EdgeDataRef;

#ifndef BRUTAL_CAST  // this works only with GNU-Compiler

	typedef EdgeDataRef EdgeDataPtr;
	typedef NodeDataRef NodeDataPtr;

	struct NeighbourData
	{
		NodeDataPtr	_n;
		EdgeDataPtr	_e;
	};

#endif //#ifndef BRUTAL_CAST
	
	// zentrale Verwaltungsinfo fr alle Attribute:
	mutable AttributeVectorTable nodeAttributeVectorTable, edgeAttributeVectorTable;

	mutable std::list<NodeData> 			nodes;
	mutable std::list<EdgeData> 			edges;
	mutable std::list<std::list<NeighbourData> >	neighbours;

// this 4 function realises brutal casts from void** to ::iterator and vice versa.
// this is necessary to avoid compiler conflicts in case of using std::list in a cyclic definition

#ifndef BRUTAL_CAST  // this works only with GNU-Compiler

static NodeDataRef Cast2NodeDataRef( NodeDataPtr x ) { return x; };
static NodeDataPtr Cast2NodeDataPtr( NodeDataRef x ) { return x; }; 
static EdgeDataRef Cast2EdgeDataRef( EdgeDataPtr x ) { return x; }; 
static EdgeDataPtr Cast2EdgeDataPtr( EdgeDataRef x ) { return x; }; 

#else

static NodeDataRef Cast2NodeDataRef( NodeDataPtr x ) { NodeDataRef ref; memcpy( &ref, &x, sizeof( ref ) ); return ref; };
static NodeDataPtr Cast2NodeDataPtr( NodeDataRef x ) { NodeDataPtr ptr; memcpy( &ptr, &x, sizeof( ptr ) ); return ptr; };
static EdgeDataRef Cast2EdgeDataRef( EdgeDataPtr x ) { EdgeDataRef ref; memcpy( &ref, &x, sizeof( ref ) ); return ref; };
static EdgeDataPtr Cast2EdgeDataPtr( EdgeDataRef x ) { EdgeDataPtr ptr; memcpy( &ptr, &x, sizeof( ptr ) ); return ptr; };

#endif //#ifndef BRUTAL_CAST 
};

// ****************************** Node ************************************

class Node : public AttributedElement
{
public:
	Node() : _g(0) {}
	Node NextNode(void) const;	
	Node PreviousNode(void) const;	
	Edge FirstIngoingEdge(void) const;
	Edge FirstOutgoingEdge(void) const;
	Edge LastIngoingEdge(void) const;
	Edge LastOutgoingEdge(void) const;
	friend bool operator==(const Node &n1, const Node &n2);
	friend bool operator!=(const Node &n1, const Node &n2);
	operator bool() const;	// valid/existing node? 

private:
	friend class Graph;
	friend class Edge;
	Node(const Graph *_g, std::list<Graph::NodeData>::iterator _n) : _g(_g), _n(_n) { SetAttributeVectorPtr(_n->attr);}
	const Graph *_g;	// =0 invalid pointer
	std::list<Graph::NodeData>::iterator _n;
};

// ****************************** NodePair ************************************

class NodePair
{
public:
	NodePair() : m_first( Node()),  m_second( Node() ) {};
	NodePair( const Node &first, const Node &second ) : m_first( first),  m_second( second ) {};
	NodePair(const NodePair &p) { m_first = p.m_first; m_second = p.m_second; };
	NodePair& operator=(const NodePair &p) { m_first = p.m_first; m_second = p.m_second; return *this; };
	operator bool() const { return m_first && m_second;	};
	Node First( void ) const { return m_first; } ;
	Node Second( void ) const { return m_second; } ;

private:
	Node m_first;
	Node m_second;
};

// ****************************** Edge ************************************

class Edge : public AttributedElement
{
public:
	Edge() : _g(0) {}
	Edge NextEdge(void) const;	
	Edge PreviousEdge(void) const;	
	Node SourceNode(void) const;
	Node TargetNode(void) const;
	Edge NextIngoingEdge(void) const;	// next edge with identical target node
	Edge PreviousIngoingEdge(void) const;	// previous edge with identical target node
	Edge NextOutgoingEdge(void) const;	// next edge with identical source node
	Edge PreviousOutgoingEdge(void) const;	// previous edge with identical source node
	friend bool operator==(const Edge &e1, const Edge &e2);
	friend bool operator!=(const Edge &e1, const Edge &e2);
	operator bool() const;	// valid/existing edge

private:
	friend class Graph;
	friend class Node;
	Edge(const Graph *_g, std::list<Graph::EdgeData>::iterator _e) : _g(_g), _e(_e) { SetAttributeVectorPtr(_e->attr);}
	const Graph *_g;	// =0 -> ungltig
	std::list<Graph::EdgeData>::iterator _e;
};


// ****************************** NodeAttribute ************************************

template<typename AttributeType> class NodeAttribute : public Attribute< Node, AttributeType>
{
public:
	NodeAttribute( const AttributeType defaultValue=AttributeType()) : Attribute< Node, AttributeType >( defaultValue ) {};
	NodeAttribute( const Attribute< Node, AttributeType >( a ) ) : Attribute< Node, AttributeType >( a ) {};
};


// ****************************** EdgeAttribute ************************************

template<typename AttributeType> class EdgeAttribute : public Attribute< Edge, AttributeType >
{
public:
	EdgeAttribute( const AttributeType defaultValue=AttributeType()) : Attribute< Edge, AttributeType >( defaultValue ) {};
	EdgeAttribute( const Attribute< Edge, AttributeType >( a ) ) : Attribute< Edge, AttributeType >( a ) {};
};

// ****************************** NodeVector ************************************

class NodeVector : public SUVector< Node > 
{
public:
	NodeVector( const int numberOfNodes = 0 ) : SUVector< Node >( numberOfNodes ) {};
	NodeVector(const NodeVector &vec) : SUVector< Node >( vec ) {};
	int NumberOfNodes( void ) const { return Size(); };
};


// ****************************** Relation ************************************

class Relation : public SUVector< NodePair >
{
public:
	int NumberOfPairs( void ) const { return Size(); };
};


// ****************************** EdgeVector ************************************

class EdgeVector : public SUVector< Edge >
{
public:
	EdgeVector( const int numberOfEdges = 0 ) : SUVector< Edge >( numberOfEdges ) {};
	EdgeVector(const EdgeVector &vec) : SUVector< Edge >( vec ) {};
	int NumberOfEdges( void ) const { return Size(); };
};


// ****************************** NodePath ************************************

class NodePath : public NodeVector 
{
public:
	NodePath( const int numberOfNodes = 0 ) : NodeVector( numberOfNodes ) {};
	NodePath(const NodePath &p) : NodeVector( p ) {};
	void Print( char *comment, NodeAttribute< int > &attribut ); // ?? delete later
};


// ****************************** EdgePath ************************************

class EdgePath : EdgeVector
{
public:
	EdgePath( const int numberOfNodes = 0 ) : EdgeVector( numberOfNodes ) {};
	EdgePath(const EdgePath &p) : EdgeVector( p ) {};
};

// ****************************** NodePathVector ************************************

class NodePathVector : public SUVector< NodePath > 
{
public:
	NodePathVector( const int numberOfPathes = 0 ) : SUVector< NodePath >( numberOfPathes ) {};
	int NumberOfPathes( void ) const { return Size(); };
};


// ****************************** EdgePathVector ************************************

class EdgePathVector : public SUVector< EdgePath >
{
public:
	EdgePathVector( const int numberOfPathes = 0 ) : SUVector< EdgePath >( numberOfPathes ) {};
	int NumberOfPathes( void ) { return Size(); };
};



#endif // #ifdef GRAPH_H

