////////////////////////////////////////////////////////////////////////////////
// File         : route.cpp
// Subsystem    : Graph
// Component    : graphs
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class route
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 03.09.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 07.01.04    Josef Huebl		 additional printf()
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////


#include "route.h"

//#define TRACE

#ifdef TRACE
#include <stdio.h>
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////
bool Router::SearchColoredPathes(  NodePathVector &pathes )
{
	
	if ( FastCheck( pathes ) == false )
	{

#ifdef TRACE
		printf( "ERROR: FastCheck() fails!\n" );
#endif //#ifdef TRACE

		return false;
	}

	SortPathesToComponents( pathes ); // sort pathes in groups according to component of first node 

	m_numberOfPathes = pathes.NumberOfPathes();
	m_currentPath = &(pathes[0]);
	m_currentPathNumber = 0;
	m_currentPathColor = m_color[ m_currentPath->First() ];
	m_noneColor = 0; // color.GetDefaultValue();
	m_mumberOfCombinations = 1; //for debugging only

   /* Save stack environment for return in case of error. First 
    * time through, jmpret is 0, so true conditional is executed. 
    * If an error occurs, jmpret will be set to -1 and false 
    * conditional will be executed.
    */

   int jmpret = setjmp( m_jmp_mark ); 

   if( jmpret == 0 ) // longjmp() will come back to this line
   {
	return ValidStep(); // this is always called
   }
   else
   {
	   // this is done, when ValidStep() is canceld by a longjmp() call
	   // which means ValidPath() founds a component, for which the colored tree problem
	   // never can be solved.

	   return false;
   }
}


bool Router::FastCheck(  NodePathVector &pathes )
{
	NodePathVector singlePath;
	Node first;
	NodeAttribute<int> color( m_color ); // save colors

	for ( int i=0; i<pathes.Size(); i++)
	{
		m_numberOfPathes = 1;
		m_currentPath = &(pathes[i]);
		m_currentPathNumber = 0;
		first = m_currentPath->First();
		m_currentPathColor = m_color[ first ];
		m_noneColor = 0; // color.GetDefaultValue();
		m_mumberOfCombinations = 1; //for debugging only

		if ( ValidStep() == false )
		{
			printf(" No valid path[ %d ]!!!!\n", i );
			return false;
		}

		// reset colors and path
		m_color = color;
		m_currentPath->DelAll();
		m_currentPath->AddLast( first );
	}

	return true;
}


void Router::SortPathesToComponents( NodePathVector &pathes ) // according to their first node
{
	Node firsti, firstj;
	int componenti, componentj;

	for (int i=0; i<pathes.Size(); i++)
	{
		firsti = pathes[ i ].First();
		componenti = m_componentNumber[ firsti ];

		for (int j=i+1; j<pathes.Size(); j++)
		{
			firstj = pathes[ j ].First();
			componentj = m_componentNumber[ firstj ];

			if ( componentj == componenti)
			{
				firsti = pathes[ ++i ].First();

				if ( j != i)
				{
					// switch firsti and firstj
					pathes[ i ].DelLast();
					pathes[ i ].AddLast( firstj );
					pathes[ j ].DelLast();
					pathes[ j ].AddLast( firsti );
				}
			}
		}
	}
}


bool Router::CheckPathes( NodeVector &sources, NodeVector &targets, NodePathVector &pathes )
{
	int i, j, k, colorFirst, node1Color;
	bool found, result = true;
	Node node, node1, node2;
	int numberOfSources = sources.NumberOfNodes();
	int numberOfTargets = targets.NumberOfNodes();

	// check targets on all pathes
	for (i=0; i<numberOfTargets; i++)
		if ( pathes[i].First() != targets[i] )
		{
			printf( "ERROR: wrong target node on path[ %d ]!\n", i );
			result = false;
		}

	//check colors on all pathes
	for (i=0; i<numberOfTargets; i++)
	{
		node = pathes[i].First();
		colorFirst = m_color[ node ];

		while ( node )
		{
			if ( m_color[ node ] != colorFirst )
			{
				printf( "ERROR: wrong color on path[ %d ]!\n", i );
				result = false;
			}

			node = pathes[i].Next();
		}
	}

	//check disjoint for all pathes
	for (i=0; i<numberOfTargets; i++)
	{
		for (j=i+1; j<numberOfTargets; j++)
		{
			node1 = pathes[i].Last();
			node1 = pathes[i].Prev(); // finial sources can be common

			while ( node1 )
			{
				node2 = pathes[j].Last();
				node2 = pathes[j].Prev();  // finial sources can be common

				while ( node2 )
				{
					if ( node1 == node2 )
					{
						printf( "ERROR: identical nodes on path[ %d ] and path[ %d ]!\n", i, j );
						result = false;
					}
					node2 = pathes[j].Prev();
				}

				node1 = pathes[i].Prev();
			}
		}
	}

	//check connectivity for all pathes
	for (i=0; i<numberOfTargets; i++)
	{
		node1 = pathes[i].Last();
		node2 = pathes[i].Prev();

		while ( node2 )
		{
			if ( !(m_graph.ExistingEdge( node1, node2 )) )
			{
				printf( "ERROR: missing edge on path[ %d ]!\n", i );
				result = false;
			}

			node1 = node2;
			node2 = pathes[i].Prev();
		}
	}

	//check SourceNode nodes
	for (i=0; i<numberOfTargets; i++)
	{
		node1 = pathes[i].Last();
		found = false;

		for (k=0; k<numberOfSources; k++)
		{
			if ( node1 == sources[k] )
			{
				found = true;
				break; // according SourceNode node found
			}
		}

		if ( !found )
		{
			// search on other pathes
			node1Color = m_color [ node1 ];

			for (j=0; j<numberOfTargets; j++)
				if (( j != i ) && ( m_color[ pathes[j].First() ] == node1Color ) )
				{
					node2 = pathes[j].First();

					while ( node2 )
					{
						if ( node1 == node2 )
						{
							found = true;
							j = numberOfTargets; //finish loop
							break;
						}

						node2 = pathes[j].Next();
					}
				}
		}

		if ( !found )
		{
			printf( "ERROR: no soure node on path[ %d ] !\n", i );
			result = false;
		}
	}


	return result;
}


//////////////////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////////////////

bool Router::ValidStep( void )
{
	Node last = m_currentPath->Last();
	Edge edge = last.FirstIngoingEdge();
	Node pre;

#ifdef TRACE
		printf("ValidStep() last = %d color =%d\n", m_id[last], m_color[last] );
#endif

	while (edge)  // for all predecessors of last
	{
		pre = edge.SourceNode();

#ifdef TRACE
		printf("ValidStep() (%d, %d)   ", m_id[edge.TargetNode()], m_id[edge.SourceNode()]);
		printf("color[ %d ] = %d\n", m_id[ pre] ,m_color[ pre ] );
#endif
		if( m_color[ pre ] != m_noneColor ) 
		{
			// SourceNode node is already colored and pathes can meet there
			if (( m_color[ pre ] == m_currentPathColor ) && m_multible[ pre ] )
			{
				// path finished
				m_currentPath->AddLast( pre );
				if ( ValidPath() )
					return true;
				else
				{
					m_currentPath->DelLast( );
				}
			}
		}
		else
		{
			// SourceNode node is uncolored
			m_currentPath->AddLast( pre );
			m_color[ pre ] = m_currentPathColor;
			if ( ValidStep() )
				return true;
			else
			{
				m_currentPath->DelLast();
				m_color[ pre ] = m_noneColor;
			}

		}
		edge = edge.NextIngoingEdge();
	}
	return false;
}


bool Router::ValidPath( void )
{
#ifdef TRACE
	printf("ValidPath() currentPathNumber = %d\n", m_currentPathNumber );
#endif

	printf("m_mumberOfCombinations = %ld\n", m_mumberOfCombinations++ );

	if ( m_currentPathNumber == (m_numberOfPathes - 1))
		return true; // no other path

	// try to finish all next pathes
	m_currentPath++;
	m_currentPathNumber++;
	m_currentPathColor = m_color[ m_currentPath->First() ];

	if ( !ValidStep() )
	{
		// check if no path in component
		if (m_componentNumber[ m_currentPath->First() ] != m_componentNumber[ (m_currentPath-1)->First() ] )
		{
			// no valid path in the current component
			longjmp( m_jmp_mark, -1 ); // jump back to all of ValidStep() in SearchColoredPathes()
		}

		// no valid pathes found
		m_currentPath--;
		m_currentPathNumber--;
		m_currentPathColor = m_color[ m_currentPath->First() ];

		return false;
	}

	return true;
}


